(function(PLUGIN_ID) {
  'use strict';
  var fileElements = document.getElementsByClassName('file'),
    submitElement = document.getElementById('submit'),
    cancelElement = document.getElementById('cancel'),
    editBtnName = document.getElementById('editBtnName'),
    editBtnCode = document.getElementById('editBtnCode'),
    attachmentCode = document.getElementById('attachmentCode'),
    initialFontSize = document.getElementById('initialFontSize'),
    initialBrushSize = document.getElementById('initialBrushSize'),
    config = kintone.plugin.app.getConfig(PLUGIN_ID); 
  if (config && (typeof config === 'object' && config !== null)) {
    // selectフィールドの初期値
    // フォントサイズ
    if (!('initialFontSize' in config)) {
      config.initialFontSize = "36";
    }
    // ブラシサイズ
    if (!('initialBrushSize' in config)) {
      config.initialBrushSize = "40";
    }
    Object.keys(config).forEach(function(key, index) {
      if(/^preview/.test(key)) {
        var img = document.getElementById('img-' + key) ? document.getElementById('img-' + key) : document.createElement('img');
        img.id = 'img-' + key;
        img.src = config[key];
        if(document.getElementById(key)) {
          document.getElementById(key).innerHTML = img.outerHTML;
        }
      } else if(document.getElementById(key)) {
        document.getElementById(key).value = config[key];
      }
    });
  } else {
    config = {};
    config.editBtnName = "画像書き込み";
    config.initialFontSize = "36";
    config.initialBrushSize = "40";
    editBtnName.value = config.editBtnName;
    initialFontSize.value = config.initialFontSize;
    initialBrushSize.value = config.initialBrushSize;
  }

  [].forEach.call(fileElements, function(fileElement, index) {
    fileElement.addEventListener('change', function(event) {
      var input = event.target,
        parentElement = input.closest('.form-group'),
        previewElement = parentElement.querySelector('.preview'),
        target = previewElement.getAttribute('id'),
        reader = new FileReader();
      
      if (!input.files.length) return;
      if(input.files[0].size > 500000) {
        alert(util.sprintf($config.invalid_image_size, 500.000));
        input.value = '';
        return false;
      }
      reader.onload = function(e) {
        var img = document.createElement('img');
        img.id = 'img-' + target;
        img.src = e.target.result;
        config[target] = e.target.result;
        previewElement.innerHTML = img.outerHTML;
      };

      reader.readAsDataURL(input.files[0]);
    }, false);
  });

  submitElement.addEventListener('click', function(event) {
    // inputフィールド→config
    var inputs = document.querySelectorAll('input.form-control');
    Array.prototype.forEach.call(inputs, function(element) {
      var k = element.getAttribute('id'),
        v = element.value;
      config[k] = v;
    });
    // selectフィールド→config
    var selects = document.querySelectorAll('select.form-control');
    Array.prototype.forEach.call(selects, function(element) {
      var k = element.getAttribute('id'),
        v = element.value;
      config[k] = v;
    });
    if (!(typeof config === 'object' && config !== null)) return;
    kintone.plugin.app.setConfig(config);
  });

  cancelElement.addEventListener('click', function(event) {
      window.location.href = '/k/admin/app/' + kintone.app.getId() + '/plugin/';
  });
  
  // アプリのレイアウト情報を取得（スペースフィールド）
  //var editBtnCode = document.getElementById('editBtnCode');
  kintone.api(kintone.api.url('/k/v1/app/form/layout', true), 'GET', {
    app: kintone.app.getId()
  }, function(resp) {
    for (var ii in resp.layout) {
      // スペースフィールド
      var layout_elem = resp.layout[ii];
      if ('ROW' == layout_elem.type) {
        for (var jj in layout_elem.fields) {
          if ('SPACER' == layout_elem.fields[jj].type) {
            var elementId = layout_elem.fields[jj].elementId;
            var option = document.createElement('option');
            option.value = elementId;
            option.textContent = elementId;
            if (typeof (config['editBtnCode']) !== 'undefined' && elementId === config['editBtnCode']) {
              option.selected = true;
            }
            editBtnCode.appendChild(option);
          }
        }
      }
      // グループ内のスペースフィールド
      else if ('GROUP' == layout_elem.type) {
        for (var kk in layout_elem.layout) {
          var group_layout_elem = layout_elem.layout[kk];
          if ('ROW' == group_layout_elem.type) {
            for (var jj in group_layout_elem.fields) {
              if ('SPACER' == group_layout_elem.fields[jj].type) {
                var elementId = group_layout_elem.fields[jj].elementId;
                var option = document.createElement('option');
                option.value = elementId;
                option.textContent = elementId;
                if (typeof (config['editBtnCode']) !== 'undefined' && elementId === config['editBtnCode']) {
                  option.selected = true;
                }
                editBtnCode.appendChild(option);
              }
            }
          }
        }
      }
    };
    // アプリのフォーム情報を取得（スペースフィールド以外）
    //var attachmentCode = document.getElementById('attachmentCode');
    
    kintone.api(kintone.api.url('/k/v1/app/form/fields', true), 'GET', {
      app: kintone.app.getId()
    }, function(resp) {
      for (var key in resp.properties) {
        if (!resp.properties.hasOwnProperty(key)) {continue;}
        var prop = resp.properties[key];
        if (prop.type === 'FILE') {
          var option = document.createElement('option');
          option.value = prop.code;
          option.textContent = prop.code;

          if (typeof (config['attachmentCode']) !== 'undefined' && prop.code === config['attachmentCode']) {
            option.selected = true;
          }
          attachmentCode.appendChild(option);
        }
      } 
    })
  });

})(kintone.$PLUGIN_ID);

function isCssIdValid(id) {
  var re = /^[A-Za-z]+[\w\-\:\.]*$/
  return re.test(id)
}