// **********************************************************************************
//
//  editable.js
//
//  履歴
//    2022.01.01    画像カキコミプラグインとしてリリース
//
//                                              Copyright(C)2022 Caminova, Inc.
// **********************************************************************************
(function() {

  // Define our constructor
  this.Simulator = function() {
    fabric.Image.prototype.getSvgSrc = function() {
      return this.toDataURLforSVG();
    };

    fabric.Image.prototype.toDataURLforSVG = function(options) {
      var el = fabric.util.createCanvasElement();
      el.width = this._element.naturalWidth || this._element.width;
      el.height = this._element.naturalHeight || this._element.height;
      el.getContext("2d").drawImage(this._element, 0, 0);
      return el.toDataURL(options);
    };
    // Create global element references
    this.btnPopup = null;
    this.popup = null;
    this.simulator = null;
    this.sidebar = null;
    this.main = null;
    this.canvas = null;
    this.canvasElement = null;
    this.colorPicker = null;
    //図形ツール
    this.controls = null;
    //テキストツール
    this.textControls = null;
    //カラー指定
    this.colorControls = null;
    //画面の拡大・縮小
    this.magControls = null;
    //オブジェクトの削除
    this.delControls = null;
    //保存・閉じる
    this.actions = null;



    this.blockControls = null;
    this.lists = null;
    this.kintone = null;
    this.imgObject = null;
    this.dragging = false;
    this.alert = null;
    this.loading = null;
    this._attachedEvents = [];
    this.output = {};
    this.outputAttr = {};
    this.scale = 1;
    this.brush = {
      color: '#000000',
      background: 'transparent',
      size: 3,
      fontSize: 8,
      maxSize: 100
    }
    this.originWith = 0;
    this.originHeight = 0;
    this.viewBox = {};

    // Define option defaults
    var defaults = {
      pluginId: '',
    };

    // Create options by extending defaults with the passed in arugments
    if (arguments[0] && typeof arguments[0] === "object") {
      this.options = extendDefaults(defaults, arguments[0]);
    }
    this.pluginId = this.options.pluginId;
    this.localStorageKey = 'storage_' + this.options.pluginId;
    //this.deviceDetect = window.localStorage.getItem(this.localStorageKey);
    //202201 nishitani deviceDetect: true ---> mobile
    this.deviceDetect = false;

    // 現在のzoomインデクス
    this.zoomindex = 0;

    // ブラウザの拡大縮小抑止
    window.addEventListener("mousewheel", mousewheelFunc, { passive: false } );
    // ctrlキー押下検出
    document.addEventListener("keydown", keyDownFunc);
    document.addEventListener("keyup", keyUpFunc);

    // 認証による実行可否
    this.runable = false;
  };

  /**
   * init
   * @param  kintone event
   * @description: apply btn toggle simulator
   */
  Simulator.prototype.init = function(event) {
    var self = this;
    self.brush = Object.assign(self.brush, self.options.brushControls);
    self.kintone = self.deviceDetect ? kintone.mobile.app : kintone.app;
    var pluginConfig = self.options.pluginConfig,
      spaceId = (pluginConfig.editBtnCode || 'simulator'),
      editBtnSpace = self.kintone.record.getSpaceElement(spaceId);
    if(!editBtnSpace) return;
    editBtnSpace.className = 'simulator-space';

    // 実行可能？（認証）
    if (!this.runable) {
      return;
    }

    // init button simulator
    self.btnPopup = util.createElement('button', {
      class: 'btn-primary caminova-btn-kakikomi',
    }, pluginConfig.editBtnName || '画像書き込み');
    editBtnSpace.appendChild(self.btnPopup);
    self._attachEvent(self.btnPopup, 'click', function() {
      self.initPopup();
    });
    
    // 添付画像がないときは「画像編集」ボタンを押せないように
    var attachmentCode = (pluginConfig.attachmentCode || '');
    //if(!target || !target.querySelectorAll('img').length) {
    //  // 添付画像なし、または、添付ファイルフィールドなし
    //  self.btnPopup.disabled = true;
    //}
    var retrycount = 0;
    var intervalid = setInterval(function() {
      var target = self.kintone.record.getFieldElement(attachmentCode);
      if(target && target.querySelectorAll('img').length) {
        clearInterval(intervalid);
        intervalid = null;
        return
      }
      //3秒（500ms x 6回）待っても画像が無いようなら、グレーアウトさせる
      retrycount++;
      if (6 <= retrycount) {
        self.btnPopup.disabled = true;
        clearInterval(intervalid);
        intervalid = null;
        return
      }
    }, 500);
  };

  /**
   * initPopup
   * @param  
   * @description: render popup
   */
  Simulator.prototype.initPopup = function() {
    var self = this,
      brushOptions = self.options.brushControls;
    var magOptions = self.options.magControls;
    // フォント、ブラシサイズの初期値
    brushOptions.fontSize = self.options.pluginConfig.initialFontSize;
    brushOptions.size = self.options.pluginConfig.initialBrushSize;
    self.brush.fontSize = brushOptions.fontSize;
    self.brush.size = brushOptions.size;
    if(self.popup) return;
    self.popup = util.createElement('div', {
      id: self.pluginId,
      class: self.deviceDetect ? 'editable-wrap is-mobile' : 'editable-wrap',
    });
    self.simulator = util.createElement('div', {
      class: 'editable-container',
    });
    self.sidebar = util.createElement('sidebar', {
      id: self.generateKey('sidebar'),
      class: 'editable-sidebar',
    });
    self.main = util.createElement('main', {
      id: self.generateKey('main'),
      class: 'editable-main',
    });
    //
    //図形ツール（スタンプ、ドロー、線の太さ）
    self.controls = util.createElement('div', {
      id: self.generateKey('control'),
      class: 'editable-control',
      style: 'border-right: 1px solid #CCC;text-align: center;'
    });
    //タイトル
    self.controls.appendChild(util.createElement('div',{
      class:"editable-title editable-title-zukei"
      //style:"font-size:smaller;padding: 0.25em;background:#EEE"
    },"図形ツール"));
    //スタンプ
    self.registerButton(self.controls, self.options.controls, self.controlEvents);
    //ドロー
    var btnDraw = util.createElement('button', {
      class: brushOptions.class + ' btn-control',
    }, brushOptions.name);
    var wrap_btnDraw = util.createElement('div', {
      class: 'btn-wrap'
    });
    wrap_btnDraw.appendChild(btnDraw);
    self.controls.appendChild(wrap_btnDraw);
    //線の太さ
    var wrap_futosa = util.createElement('div', {
      class: 'btn-wrap'
    });
    var title_futosa = util.createElement('div', {
      style: 'font-size:small;padding-top:0.25em;'
    },"線の太さ");
    wrap_futosa.appendChild(title_futosa);
    wrap_futosa.appendChild(util.dropdown(
      brushOptions.sizes,
      { defaultLabel: brushOptions.brushSizeName, defaultValue: brushOptions.size },
      function(size) {
        self.brush.size = size;
        self.updateActiveItems();
      }
      )
    );
    self.controls.appendChild(wrap_futosa);
    //
    //テキストツール（テキストボックス作成、文字サイズ）
    self.textControls = util.createElement('div', {
      id: self.generateKey('textControl'),
      class: 'editable-control',
      style: 'border-right: 1px solid #CCC;text-align: center;align-items:inherit;'
    });
    //タイトル
    self.textControls.appendChild(util.createElement('div',{
      class:"editable-title editable-title-text"
      //style:"font-size:smaller;padding: 0.25em;background:#EEE"
    },"テキストツール"));
    //テキストボックスの作成
    var wrap_textbox = util.createElement('div', {
      class: 'btn-wrap',
      style: 'margin-right:0.25rem;'
    });
    var btnText = util.createElement('button', {
      class: brushOptions.textClass + ' btn-control',
    }, brushOptions.textName);
    wrap_textbox.appendChild(btnText);
    self.textControls.appendChild(wrap_textbox);
    var wrap_textsize = util.createElement('div', {
      class: 'btn-wrap'
    });
    var title_textsize = util.createElement('div', {
      style: 'font-size:small;padding-top:0.25em;'
    },"文字サイズ");
    wrap_textsize.appendChild(title_textsize);
    wrap_textsize.appendChild(util.dropdown(
      brushOptions.fontSizes,
      { defaultLabel: brushOptions.fontSizeName, defaultValue: brushOptions.fontSize },
      function(fontSize) {
        self.brush.fontSize = fontSize;
        self.updateActiveItems();
      }
      )
    );
    self.textControls.appendChild(wrap_textsize);
    //
    //カラー指定
    self.colorControls = util.createElement('div', {
      id: self.generateKey('colorControl'),
      class: 'editable-control',
      style: 'border-right: 1px solid #CCC;text-align: center;align-items:inherit;'
    });
    //タイトル
    self.colorControls.appendChild(util.createElement('div',{
      class:"editable-title editable-title-color"
      //style:"font-size:smaller;padding: 0.25em;background:#EEE"
    },"カラー指定"));
    //文字色
    self.colorControls.appendChild(util.registerColorPicker(brushOptions, function(color) { 
      self.brush.color = color;
      self.canvas.freeDrawingBrush.color = self.brush.color;
      self.updateActiveItems();
    }));
    //背景色
    self.colorControls.appendChild(util.registerColorPicker(Object.assign(brushOptions, { color: 'transparent', fill: true }), function(color) {
      self.brush.background = color;
      self.updateActiveItems();
    }));
    //
    //画面の表示倍率
    self.magControls = util.createElement('div', {
      id: self.generateKey('magControl'),
      class: 'editable-control',
      style: 'border-right: 1px solid #CCC;text-align: center;align-items:inherit;'
    });
    //タイトル
    self.magControls.appendChild(util.createElement('div',{
      class:"editable-title editable-title-mag"
      //style:"font-size:smaller;padding: 0.25em;background:#EEE"
    },"表示倍率"));
    //拡大・縮小
    self.registerButton(self.magControls, self.options.magControls, self.controlEvents);
    //
    //オブジェクトの削除
    self.delControls = util.createElement('div', {
      id: self.generateKey('delControl'),
      class: 'editable-control',
      style: 'border-right: 1px solid #CCC;text-align: center;align-items:inherit;'
    });
    //タイトル
    self.delControls.appendChild(util.createElement('div',{
      class:"editable-title editable-title-del"
      //style:"font-size:smaller;padding: 0.25em;background:#EEE"
    },"オブジェクトの削除"));
    //削除
    self.registerButton(self.delControls, self.options.delControls, self.controlEvents);
    //
    //保存、閉じる
    self.actions = util.createElement('div', {
      id: self.generateKey('action'),
      class: 'editable-control editable-action'
    });
    self.registerButton(self.actions, self.options.actions, self.actionEvents);

    self.lists = util.createElement('ul', {
      class: 'editable-thumbs loading',
    });
    self.canvasElement = util.createElement('canvas', {
      id: self.generateKey('canvas'),
      class: 'editable-simulator'
    });
    util.appendElements(self.sidebar, [
      util.registerBlock([self.controls, self.textControls, self.colorControls, self.magControls, self.delControls, self.actions])
    ]);
    util.appendElements(self.simulator, [self.sidebar, self.lists, self.main]);
    self.main.appendChild(self.canvasElement);
    self.popup.appendChild(self.simulator);
    
    self.registerLists();

    // 画像ロード終わってから編集画像を開く
    // self.initDefaultCanvas();
    self._attachEvent(btnText, 'click', self.controlEvents);
    self._attachEvent(btnDraw, 'click', self.registerDrawEvent);
    document.querySelector('body').appendChild(self.popup);
  };

  Simulator.prototype.registerLists = function() {
    var self = this, 
      // target = this.deviceDetect ? document.querySelector('.control-showlayout-file-gaia') : document.getElementById('record-gaia'),
      attachmentCode = (self.options.pluginConfig.attachmentCode || ''),
      target = self.kintone.record.getFieldElement(attachmentCode);
    
    if(!target || !target.querySelectorAll('img').length) {
      Object.keys(self.options.pluginConfig).forEach(function(key, index) {
        if(self.options.pluginConfig[key] && /^preview/.test(key)) {
          self.makeList(self.options.pluginConfig[key], index);
        }
      });
    } else {
      var imgs = target.querySelectorAll('img');
      for(var i = 0; i < imgs.length; i++) {
        var url = imgs[i].src;
        self.makeList(url, i);
        self.output[i] = url;
      }
    }
    
    self.lists.classList.remove('loading');
  };

  Simulator.prototype.makeList = function(url, index) {
    var self = this; 
    var img = util.createElement('img', {
      src: url
    });
    index = !index ? 0 : index;
    var li = util.createElement('li', {
      class: 'editable-item',
      index: index
    });
    this._attachEvent(li, 'click', this.listEvents);
    li.appendChild(img);
    this.lists.appendChild(li);
    // 1枚目なら編集画像を開く
    if (0 == index) {
      self.initDefaultCanvas();
    }
  };

  Simulator.prototype.makeList2 = function(url, alt, index) {
    var self = this; 
    var img = util.createElement('img', {
      src: url
    });
    index = !index ? 0 : index;
    //var li = util.createElement('li', {
    //  class: 'editable-item',
    //  index: index
    //});
    //this._attachEvent(li, 'click', this.listEvents);
    // index番目のli要素を探す
    var li = this.lists.querySelector('li[index="' + index + '"]');
    li.appendChild(img);
    //this.lists.appendChild(li);
    // イメージのファイル名、外部アプリID、レコードIDを保存する
    this.outputAttr[index] = alt;
    // 1枚目なら編集画像を開く
    if (0 == index) {
      self.initDefaultCanvas();
    }
  };

  Simulator.prototype.initDefaultCanvas = function() {
    if(!this.lists.querySelector('li:first-child')) return;
    var li = this.lists.querySelector('li:first-child');
    li.classList.add('active');
    var url = li.querySelector('img').src;
    if(util.isSvgFromStringUrl(url)) {
      this.loadSVGFromUrl(li.querySelector('img').src);
    } else {
      this.loadImgFromUrl(li.querySelector('img').src);
    }
  };

  Simulator.prototype.loadImgFromUrl = function(url) {
    var self = this;
    url = removeParameterFromUrl(url, ['w', 'h', 'flag', 'thumbnail']);
    self.createLoading('image-from-url');
    fabric.Image.fromURL(url, function(img) {
      self.initCanvas(img, self);
    });
  };

  Simulator.prototype.loadSVGFromUrl = function(url) {
    var self = this;
    url = removeParameterFromUrl(url, ['w', 'h', 'flag', 'thumbnail']);
    fabric.loadSVGFromURL(url, function(objects, options) {
      var img = new fabric.util.groupSVGElements(objects, options);
      self.initCanvas(img, self);
    });
  };

  Simulator.prototype.initCanvas = function(img, self) {
    img.selectable = false;
    img.id = 'item';
    self.originWith = img.width;
    self.originHeight = img.height;
    self.viewBox = {
      x: 0,
      y: 0,
      width: img.viewBoxWidth,
      height: img.viewBoxHeight
    };
    var scale = Math.min(
      self.main.clientWidth / img.width,
      self.main.clientHeight / img.height
    );
    self.scale = scale;
    self.zoomindex = 0;
    if(self.canvas) self.canvas.dispose();
    fabric.devicePixelRatio = 1.5;
    self.canvas = new fabric.Canvas(self.generateKey('canvas'), {
      width: img.width * scale,
      height: img.height * scale
    });
    self.canvas.on('mouse:wheel', function(e) { self.mousewheel(e); });
    var btn = document.querySelector('.btn-control.active');
    if(btn && btn.classList.contains('icon-draw')) {
      self.canvas.isDrawingMode = true;
      self.canvas.freeDrawingBrush.color = self.brush.color;
      self.canvas.freeDrawingBrush.width = parseInt(self.brush.size, 10) || 1;
      self.canvas.selection = false;
    } else {
      self.canvas.isDrawingMode = false;
      self.canvas.selection = true;
    }
    self.canvas.setZoom(scale);
    fabric.Object.prototype.objectCaching = false;
    // fabric.Object.prototype.hasControls = true;
    self.imgObject = img;

    self.canvas.add(img);
    self.initObjectEvents(img);
    self.canvas.renderAll.bind(self.canvas);
    self.removeLoading('image-from-url');
    self.registerCanvasEvents();
  };

  Simulator.prototype.registerCanvasEvents = function() {
    var self = this;
    this.canvas.on('selection:cleared', function(event) {
      console.log('selection:cleared')
    });
    this.canvas.on('before:transform', function(event) {
      console.log('before:transform')
    });
  }

  Simulator.prototype.listEvents = function(event, self) {
    self.createLoading();
    var oldTarget = document.querySelector('.editable-item.active'),
      oldIndex = oldTarget ? oldTarget.getAttribute('index') : null,
      newTarget = event.currentTarget;
    
    if(oldTarget) {
      self.renderImage(oldTarget);
      oldTarget.classList.remove('active');
    }
    newTarget.classList.add('active');
    var img = newTarget.querySelector('img'),
      url = removeParameterFromUrl(img.src, ['w', 'h', 'thumbnail']),
      newIndex = newTarget.getAttribute('index');
    var mimeType = url.match(/[^:]\w+\/[\w-+\d.]+(?=;|,)/);

    if(mimeType && mimeType[0] === 'image/svg+xml') {
      self.loadSVGFromUrl(url);
    } else {
      self.loadImgFromUrl(url);
    }
    // self.loadImgFromUrl(url);
    self.removeLoading();
  };

  Simulator.prototype.actionEvents = function(event, self) {
    if(event.target.classList.contains('btn-close')) {
      if (!self.canvas) return self.close();
      if(self.canvas.getObjects().length === 1) return self.close();
      self.alertShow().then(function(event) {
        if(event.value) {
          self.close();
        }
      });
    }
    var img = self.lists.querySelector('.editable-item.active img');
    if(img && event.target.classList.contains('btn-save')) {
      //// 編集用ワークのイメージをいったん消去する
      //self.output = [];
      self.renderImage(self.lists.querySelector('.editable-item.active'));
      self.save(event, self);
    } 
  };

  Simulator.prototype.save = function(event, self) {
    var outputs = self.output;
    self.createLoading();

    var promise = Object.keys(outputs).map(function(key, index) {
      var url = self.output[key];
      url = removeParameterFromUrl(url, ['w', 'h', 'thumbnail']);
      return fetch(url).then(function(res) {
        return res.blob(); 
      });
    });
    Promise.all(promise).then(function(resp) {
      return Promise.all(resp.map(function(blob) {
        console.log(blob)
        var fn = String(self.kintone.record.getId()) + ".jpg";
        return util.uploadFile(blob, fn);
      }));
    }).then(function(file) {
      var record = {};
      record[self.options.pluginConfig.attachmentCode] = {
        "value": file
      }
      return util.request({
        url: '/k/v1/record',
        method: 'PUT',
        body: {
          "app": self.kintone.getId(),
          "id": self.kintone.record.getId(),
          "record": record
        }
      });
    }).finally(function(success) {
      self.removeLoading();
      location.reload();
    });
  };

  Simulator.prototype.renderImage = function(target) {
    if(!target) return;
    var self = this;
    // 拡大縮小を元に戻す
    self.canvas.setZoom(self.scale);
    self.canvas.absolutePan(new fabric.Point(0,0));
    self.zoomindex = 0;

    //var oldZoom = self.canvas.getZoom();
    var oldZoom = self.scale;
    self.canvas.renderOnAddRemove = false;
    self.canvas.setZoom(1);
    var b64Data = self.canvas.toDataURL({ // see http://fabricjs.com/docs/fabric.Canvas.html#toDataURL
      format: 'jpeg',
      quality: 0.3,
      width: self.originWith,
      height: self.originHeight,
    });
    //console.log(b64Data);
    // var svg = self.canvas.toSVG({
    //   width: self.originWith,
    //   height: self.originHeight,
    //   viewBox: self.viewBox
    // });
    // console.log(self.viewBox)
    // var b64Data = 'data:image/svg+xml;base64,' + window.btoa(window.unescape(encodeURIComponent(svg)));
    self.canvas.renderOnAddRemove = true;
    self.canvas.setZoom(oldZoom);
    self.canvas.renderAll.bind(self.canvas);
    self.canvas.selection = true;
    var img = target.querySelector('img');
    img.src = b64Data;
    self.output[target.getAttribute('index')] = b64Data;
  };

  Simulator.prototype.controlEvents = function(event, self) {
    self.canvas.isDrawingMode = false; 
    self.canvas.selection = true;
    if(event.currentTarget.classList.contains('icon-remove')) {
      return self.removeObjectEvent(event);
    }
    if(event.currentTarget.classList.contains('active')) {
      
      event.currentTarget.classList.remove('active');
    } else {
      if(document.querySelector('.btn-control.active')) {
        document.querySelector('.btn-control.active').classList.remove('active');
      }
      event.currentTarget.classList.add('active');
    }
    var control = document.querySelector('.btn-control.active');

    if(control && control.classList.contains('icon-draw')) {
      self.canvas.isDrawingMode = true;
      self.canvas.freeDrawingBrush.color = self.brush.color;
      self.canvas.freeDrawingBrush.width = parseInt(self.brush.size, 10) || 1;
      self.canvas.selection = false;
    }
    // self.canvas.renderAll.bind(self.canvas);
    if(self.brushSize) {
      self._attachEvent(self.brushSize, 'change', function(event) {
        self.canvas.freeDrawingBrush.width = parseInt(event.target.value, 10) || 1;
      });
    }
  };

  Simulator.prototype.removeObjectEvent = function(event) {
    var itemActive = this.canvas.getActiveObjects();
    if (!itemActive) return;
    for(var i = 0; i < itemActive.length; i++) {
      this.canvas.remove(itemActive[i]);
    }
    this.canvas.discardActiveObject().renderAll();
  };

  Simulator.prototype.initObjectEvents = function(obj) {
    var self = this;
    obj.on('mousedown', function(event) {
      self.dragging = true;
      self.canvas.sendToBack(obj);
    });
    obj.on('mouseup', function(event) {
      if(!document.querySelector('.btn-control.active') 
        || (document.querySelector('.btn-control.active') && document.querySelector('.btn-control.active').classList.contains('icon-draw'))) return;
      if(self.dragging) {
        self.dragging = false;
        var pointer = self.canvas.getPointer(event.e);
        var control = document.querySelector('.btn-control.active');
        if(!control) return;
        // 拡大縮小ボタン？
        if (control.classList.contains('icon-mag')) {
          //console.log(control.classList);
          self.magnifyPointer(control, pointer);
        } else {
        control.classList.contains('icon-text-format') 
          ? self.addTextPointer(control, pointer)
          : self.addImagePointer(control, pointer);
        }
        event.target.setCoords();
      }
      self.canvas.sendToBack(obj);
    });
    obj.on('mousemove', function(event) {
      self.dragging = false;
    });
  };

  Simulator.prototype.addTextPointer = function(control, pointer) {
    var self = this;
    var newText = new fabric.IText("");
    newText.objectType = 'text';
    newText.left = pointer.x;
    newText.top = pointer.y;
    newText.stroke = 'transparent';
    newText.fill = self.brush.color;
    newText.fontSize = self.brush.fontSize;
    newText.backgroundColor = self.brush.background;
    newText.strokeWidth = parseInt(self.brush.size, 10);
    
    // newText.scale = this.scale;
    this.canvas.add(newText);
    this.canvas.setActiveObject(newText);
    newText.enterEditing();
    newText.selectAll();
  };

  /**
   * @description: add pattern image to canvas
   * @params: patternSize: The pattern scale will not change when the background image size changes
   */
  Simulator.prototype.addImagePointer = function(control, pointer) {
    var index = control.getAttribute('index'),
      self = this,
      controls = self.options.controls,
      bi = controls[index].icon,
      patternSize = self.options.patternSize || 60,
      img = self.lists.querySelector('li.active img'),
      unit = util.brightnessRGB(util.getAverageRGB(img));
    // if most of the background is dark
    if(unit < 5) {
      var style = control.currentStyle || window.getComputedStyle(control, false);
      bi = style.backgroundImage.slice(4, -1).replace(/"/g, "");
    }
    var keepTheRatio = function(img) {
      var min = Math.min(
          patternSize / img.width,
          patternSize / img.height
        ),
        zoom = self.canvas.getZoom(),
        scale = min / zoom;

      img.scaleX = scale;
      img.scaleY = scale;
      img.selectable = true;
      img.top = pointer.y - patternSize / (2 * zoom);
      img.left = pointer.x - patternSize / (2 * zoom);
      img.objectType = 'pattern';
      self.canvas.add(img);
      img.bringToFront();
      self.canvas.setActiveObject(img);
      self.canvas.sendToBack(self.imgObject);
      return img;
    }
    if(util.isSvgFromStringUrl(bi)) {
      fabric.loadSVGFromURL(bi, function(objects, options) {
        keepTheRatio(new fabric.util.groupSVGElements(objects, options));
      });
    } else {
      fabric.Image.fromURL(bi, function(img) {
        keepTheRatio(img);
      });
    }
  }

  Simulator.prototype.registerDrawEvent = function(event, self) {
    if(document.querySelector('.btn-control.active:not(.icon-draw)')) {
      document.querySelector('.btn-control.active:not(.icon-draw)').classList.remove('active');
    }
    if(event.target.classList.contains('active')) {
      event.target.classList.remove('active');
      self.canvas.isDrawingMode = false;
      self.canvas.selection = true;
    } else {
      event.target.classList.add('active');
      self.canvas.isDrawingMode = true;
      self.canvas.freeDrawingBrush.color = self.brush.color;
      self.canvas.freeDrawingBrush.width = parseInt(self.brush.size, 10) || 1;
      self.canvas.selection = false;
    }
  }

  Simulator.prototype.updateActiveItems = function(options) {
    var self = this;
    var activeItems = self.canvas.getActiveObjects();
    if(activeItems.length) {
      [].forEach.call(activeItems, function(activeItem) {
        var type = activeItem.objectType;
        switch(type) {
          case 'pattern':
            break;
          case 'text':
            activeItem.stroke = 'transparent';
            activeItem.fill = self.brush.color;
            activeItem.backgroundColor = self.brush.background;
            activeItem.fontSize = parseInt(self.brush.fontSize, 10);
            break;
          default:
            activeItem.stroke = self.brush.color;
            activeItem.strokeWidth = parseInt(self.brush.size, 10);
            break;
        }
        activeItem.setCoords();
      });
      self.canvas.renderAll();
    }
  }

  Simulator.prototype.close = function() {
    this.popup.remove();
    this.canvas = null;
    this.popup = null;
    this._detachEvent();
  };

  Simulator.prototype.setDevice = function(device) {
    this.deviceDetect = ('mobile' == device);
    return Boolean(device);
  };

  Simulator.prototype._attachEvent = function(el, ev, fn) {
    if(!el) return;
    var self = this;
    el.addEventListener(ev, function(event) {
      return fn(event, self);
    }, false);
    this._attachedEvents.push([el, ev, fn]);
  };

  Simulator.prototype._detachEvent = function(el, ev, fn) {
    for (var i = this._attachedEvents.length - 1; i >= 0; i--) {
      this._attachedEvents[i][0].removeEventListener(this._attachedEvents[i][1], this._attachedEvents[i][2], false);
      this._attachedEvents.splice(i, 1);
    }
  };

  Simulator.prototype.generateKey = function(prefix) {
    return prefix + '_' + this.pluginId;
  };

  Simulator.prototype.registerButton = function(el, arr, event) {
    if(!Array.isArray(arr)) return;
    var self = this;
    [].forEach.call(arr, function(arg, index) {
      var wrap = util.createElement('div', {
        class: 'btn-wrap'
      });
      var btn = util.createElement('button', {
        class: arg.class + ' btn-control' || 'btn-control',
        style: arg.style + '' || '',
        type: 'button',
        index: index
      }, arg.name);
      if (btn.classList.contains('has-nbsp')) {
        var nbsp = util.createElement('div', {
          style: 'font-size:small;padding-top:0.25em;'
        },'　');
        wrap.appendChild(nbsp);
      }
      wrap.appendChild(btn);
      el.appendChild(wrap);
      if(btn.classList.contains('icon-draw')) {
        self.brushSize = util.createElement('select', {
          class: 'editable-brush-size'
        });
        var maxSize = self.brush.maxSize;
        for(var i = 1; i <= maxSize; i++) {
          self.brushSize.appendChild(
            util.createElement('option', {
              value: i
            }, i)
          );
        }
        wrap.appendChild(self.brushSize);
      }
      self._attachEvent(btn, 'click', event);
    });
  };

  Simulator.prototype.alertShow = function() {
    var self = this;
    self.alertDestroy();
    return new Promise(function(resolve, reject) {
      var alert = document.createElement('div');
      self.alert = util.createElement('div', {
        id: self.generateKey('alert'),
        class: 'editable-alert'
      });
      var alertContainer = util.createElement('div', {
          class: 'editable-alert-container'
        }),
        alertBody = util.createElement('div', {
          class: 'editable-alert-body'
        }, self.options.alert_confirm_switch_img),
        alertAction = util.createElement('div', {
          class: 'editable-alert-action'
        }),
        btnOk = util.createElement('button', {
          class: 'btn-primary editable-alert-ok'
        }, self.options.alert_button_ok),
        btnCancel = util.createElement('button', {
          class: 'editable-alert-cancel btn-default'
        }, self.options.alert_button_cancel);
      util.appendElements(alertAction, [btnOk, btnCancel]);
      util.appendElements(alertContainer, [alertBody, alertAction]);
      self.alert.appendChild(alertContainer);
      self.sidebar.appendChild(self.alert);
      self._attachEvent(btnCancel, 'click', function(event) {
        self.alertDestroy(); 
        event.value = false;
        resolve(event); 
      });
      self._attachEvent(btnOk, 'click', function(event) {
        self.alertDestroy(); 
        event.value = true;
        resolve(event); 
      });
    });
  };

  Simulator.prototype.alertDestroy = function() {
    if(this.alert) {
      // this.canvas.clear();
      this.alert.remove();
      this._detachEvent(this.alert, 'click', null);
    }
  };

  Simulator.prototype.createLoading = function(key) {
    var id = this.generateKey(key ? 'loading-' + key : 'loading');
    if(document.getElementById(id)) return;
    var loading = util.createElement('div', {
      class: 'editable-loading loading',
      id: id
    });
    this.simulator.appendChild(loading);
  };

  Simulator.prototype.removeLoading = function(key) {
    var id = this.generateKey(key ? 'loading-' + key : 'loading');
    if(!document.getElementById(id)) return;
    document.getElementById(id).remove();
    this.loading = null;
  };

  function isMobile() {
    return (
      navigator.userAgent.match(/Android/i)
      || navigator.userAgent.match(/BlackBerry/i)
      || navigator.userAgent.match(/iPhone|iPad|iPod/i)
      || navigator.userAgent.match(/iPhone/i)
      || navigator.userAgent.match(/Opera Mini/i)
      || navigator.userAgent.match(/IEMobile/i)
    );
  }

  function removeParameterFromUrl(url, parameter) {
    var arr = Array.isArray(parameter) ? parameter : [parameter];
    [].forEach.call(arr, function(arg) {
      url = url
        .replace(new RegExp('[?&]' + arg + '=[^&#]*(#.*)?$'), '$1')
        .replace(new RegExp('([?&])' + arg + '=[^&]*&'), '$1');
    });
    return url;
  }

  function extendDefaults(source, properties) {
    var property;
    for (property in properties) {
      if (properties.hasOwnProperty(property)) {
        source[property] = properties[property];
      }
    }
    return source;
  }

  var ctrlFlg = false;          // ctrl key pressed?
  function keyDownFunc(e) {
    ctrlFlg = e.ctrlKey;        // true if ctrl key pressed.
  }
  function keyUpFunc(e) {
    ctrlFlg = e.ctrlKey;        // true if ctrl key pressed.
  }
  // ブラウザの拡大縮小抑止
  function mousewheelFunc(e) {
    if (ctrlFlg){
        if (e.preventDefault()){
            e.preventDefault();   
        } else {
            return false;
        }
    }
  }
  // 画像の拡大縮小
  Simulator.prototype.mousewheel = function(e) {
    var self = this;
    if (ctrlFlg) {              // ctrl key pressed?
      const mousex = e.pointer.x;
      const mousey = e.pointer.y;
      console.log('wheel: ' + mousex + '-' + mousey);
      const deltay = e.e.wheelDeltaY;
      var newscale = self.canvas.getZoom() + deltay / 1200;
      // 初期状態（画面いっぱい）より縮小しないようにする
      if (newscale < self.scale) {
        self.canvas.setZoom(self.scale);
        self.canvas.absolutePan(new fabric.Point(0,0));
      }
      else {
        self.canvas.zoomToPoint(new fabric.Point(mousex, mousey), newscale);
      }
    }
  }
  Simulator.prototype.magnifyPointer = function(control, pointer) {
    var self = this;
    // 移動
    if (control.classList.contains('icon-magscrl')) {

    }
    // 拡大縮小
    else {
      var mousex = pointer.x * self.scale;
      var mousey = pointer.y * self.scale;
      //console.log('button: ' + mousex + '-' + mousey);
      var deltay = (control.classList.contains('icon-magplus')) ? 1 : (-1);
      // 既定のzoomスケール
      var scalearr = [1,1.1,1.25,1.5,1.75,2,2.5,3,4,5];
      self.zoomindex += deltay;
      if (0 > self.zoomindex) {
        self.zoomindex = 0;
      }
      else if (scalearr.length <= self.zoomindex) {
        self.zoomindex = scalearr.length-1;
      }
      var newscale = self.scale * scalearr[self.zoomindex];

      // 拡大なしでスケール、表示位置を元に戻す
      if (0 == self.zoomindex) {
        self.canvas.setZoom(self.scale);
        self.canvas.absolutePan(new fabric.Point(0,0));
      }
      else {
        self.canvas.zoomToPoint(new fabric.Point(mousex, mousey), newscale);
      }
    }
  }
  // 実行可否
  Simulator.prototype.runOk = function(okng) {
    this.runable = okng;
  }

}());
