// **********************************************************************************
//
//  util.js
//
//  履歴
//    2022.01.01    画像カキコミプラグインとしてリリース
//
//                                              Copyright(C)2022 Caminova, Inc.
// **********************************************************************************
"use strict";
var util = {
  createElement: function(target, args, content) {
    var el = document.createElement(target);
    var options = Object.assign({}, args);
    Object.keys(options).map(function(key, index) {
      var value = options[key];
      el.setAttribute(key, value);
      return el;
    });
    if(content) {
      el.innerText = content;
    }
    return el;
  },

  appendElements: function(el, arr) {
    if(!(el instanceof HTMLElement) && !Array.isArray(arr)) return el;
    [].forEach.call(arr, function(arg) {
      if ((arg instanceof HTMLElement)) el.appendChild(arg);
    });
    return el;
  },

  dropdown: function(arr, option, callback) {
    var self = this;
    var defaultOption = Object.assign({
      defaultValue: '',
      defaultLabel: '',
    }, option);
    var dropdown = this.createElement('div', {
        class: 'editable-dropdown',
        width: 120
      }),
      input = this.createElement('input', {
        class: 'editable-form-control',
        placeholder: defaultOption.defaultLabel,
        value: defaultOption.defaultValue
      }),
      items = this.createElement('div', {
        class: 'editable-dropdown-items'
      });
    items.appendChild(util.createElement('div', {
      class: 'editable-dropdown-item label'
    }, defaultOption.defaultLabel));
    [].forEach.call(arr, function(name, value) {
      var cls = name == defaultOption.defaultValue ? 'editable-dropdown-item active' : 'editable-dropdown-item';
      var obj = {
        class: cls,
        data: value
      };
      var item = self.createElement('div', obj, name); 
      items.appendChild(item);
      item.onclick = function() {
        input.value = name;
        if(document.querySelector('.editable-dropdown-item.active')) {
          document.querySelector('.editable-dropdown-item.active').classList.remove('active');
        }
        
        dropdown.classList.remove('active');
        item.classList.add('active');
        if(typeof callback === 'function') {
          callback(input.value);
        }
      }
    });
    input.onfocus = function() {
      dropdown.classList.add('active');
    }
    input.onchange = function(event) {
      callback(event.target.value);
    }
    self.clickoutside(dropdown, function() {
      dropdown.classList.remove('active'); 
    });
    return self.appendElements(dropdown, [input, items]);
  },

  clickoutside: function(target, callback) {
    document.addEventListener('click', function(event) {
      var isClickInside = target.contains(event.target);
      if (!isClickInside) {
        return typeof callback === 'function' ? callback(event) : callback;
      }
    });
  },

  registerBlock: function(arr) {
    return this.appendElements(util.createElement('div', {
      class: 'editable-block'
    }), arr);
  },

  isMobile: function() {
    return {
      android: function() {
        return navigator.userAgent.match(/Android/i);
      },
      blackBerry: function() {
        return navigator.userAgent.match(/BlackBerry/i);
      },
      iOS: function() {
        return navigator.userAgent.match(/iPhone|iPad|iPod/i);
      },
      iPhone: function() {
        return navigator.userAgent.match(/iPhone/i);
      },
      opera: function() {
        return navigator.userAgent.match(/Opera Mini/i);
      },
      windows: function() {
        return navigator.userAgent.match(/IEMobile/i);
      },
      any: function() {
        return this.isMobile().android()
        || this.isMobile().blackBerry()
        || this.isMobile().iOS()
        || this.isMobile().opera()
        || this.isMobile().windows();
      }
    };
  },

  b64toBlob: function(b64Data, contentType, sliceSize) {
    var byteCharacters = atob(b64Data);
    var byteArrays = [];

    for (var offset = 0; offset < byteCharacters.length; offset += sliceSize) {
      var slice = byteCharacters.slice(offset, offset + sliceSize);

      var byteNumbers = new Array(slice.length);
      for (var i = 0; i < slice.length; i++) {
        byteNumbers[i] = slice.charCodeAt(i);
      }

      var byteArray = new Uint8Array(byteNumbers);
      byteArrays.push(byteArray);
    }

    var blob = new Blob(byteArrays, {type: contentType});
    return blob;
  },

  blobToURL: function(blob) {
    var url = window.URL || window.webkitURL;
    return url.createObjectURL(blob);
  },


  uploadFile: function(blob, fn) {
    return new Promise(function(resolve, reject) {
      // var blob = new Blob(["Sample Test File"], {type:"text\/plain"});
      var formData = new FormData();
      formData.append("__REQUEST_TOKEN__", kintone.getRequestToken());
      formData.append("file", blob, fn);

      var url = '/k/v1/file.json';
      var xhr = new XMLHttpRequest();
      xhr.open('POST', url);
      xhr.setRequestHeader('X-Requested-With', 'XMLHttpRequest');
      xhr.onload = function() {
        if (xhr.status === 200) {
          // success
          resolve(JSON.parse(xhr.responseText));
        } else {
          // error
          reject(JSON.parse(xhr.responseText));
        }
      };
      xhr.send(formData);
    });
  },

  request: function(options) {
    return new Promise(function(resolve, reject) {
      var defaultOptions = Object.assign({
        method: 'get',
        body: {},
        url: '', // '/k/v1/record'
      }, options);
      var url = kintone.api.url(defaultOptions.url, true),
        method = defaultOptions.method.toLocaleUpperCase();
      kintone.api(url, method, defaultOptions.body, function(resp) {
        resolve(resp);
      }, function(error) {
        //reject(resp);
        reject(error);
      });
    });
  },

  sprintf: function() {
    var args = arguments,
      string = args[0],
      i = 1;
    return string.replace(/%((%)|s|d)/g, function(m) {
      // m is the matched format, e.g. %s, %d
      var val = null;
      if (m[2]) {
        val = m[2];
      } else {
        val = args[i];
        // A switch statement so that the formatter can be extended. Default is %s
        switch (m) {
          case '%d':
            val = parseFloat(val);
            if (isNaN(val)) {
              val = 0;
            }
            break;
        }
        i++;
      }
      return val;
    });
  },

  getAverageRGB: function(imgEl) {
    var blockSize = 5, // only visit every 5 pixels
      defaultRGB = {r: 0, g: 0, b: 0}, // for non-supporting envs
      canvas = document.createElement('canvas'),
      context = canvas.getContext && canvas.getContext('2d'),
      data, width, height,
      i = -4,
      length,
      rgb = {r: 0, g: 0, b: 0},
      count = 0;

    if (!context) {
      return defaultRGB;
    }

    height = canvas.height = imgEl.naturalHeight || imgEl.offsetHeight || imgEl.height;
    width = canvas.width = imgEl.naturalWidth || imgEl.offsetWidth || imgEl.width;

    context.drawImage(imgEl, 0, 0);

    try {
      data = context.getImageData(0, 0, width, height);
    } catch(e) {
      /* security error, img on diff domain */
      return defaultRGB;
    }

    length = data.data.length;

    while ((i += blockSize * 4) < length) {
      ++count;
      rgb.r += data.data[i];
      rgb.g += data.data[i + 1];
      rgb.b += data.data[i + 2];
    }

    // ~~ used to floor values
    rgb.r = ~~(rgb.r / count);
    rgb.g = ~~(rgb.g / count);
    rgb.b = ~~(rgb.b / count);

    return rgb;
  },
  
  brightnessRGB: function(obj) {
    return 0.2126 * obj.r + 0.7152 * obj.g + 0.0722 * obj.b;
  },

  registerColorPicker: function(options, callback) {
    var self = this;
    var controls = Object.assign({
      fill: false,
      color: '#333',
    }, options);
    var canvasContainer = self.createElement('div', {
        class: 'editable-canvas-container',
        style: 'display: inline-block;align-items: baseline;'
      }),
      label = self.createElement('div', {
        style: 'font-size:small;padding-top:0.25em;'
      }, controls.fill ? '背景色' : '文字色'),
      picker = self.createElement('div', {
        class: controls.fill ? 'editable-colorpicker fill' : 'editable-colorpicker',
        style: 'border-color: ' + controls.color 
      }, 'A'),
      block = self.createElement('div', {
        class: 'editable-picker-block'
      });
    self.appendElements(canvasContainer, [label, picker]);
    picker.onclick = function() {
      var canvasEl = self.createElement('canvas');
      block.innerHTML = '';
      self.appendElements(block, [self.makeColorDefaults(function(color) {
        picker.setAttribute('style', 'border-color: ' + color);
        block.remove(); 
        if(typeof callback === 'function') callback(color);
      }), canvasEl]);
      canvasContainer.appendChild(block);
      self.initColorPicker(canvasEl, function(color) {
        picker.setAttribute('style', 'border-color: ' + color);
        if(typeof callback === 'function') callback(color);
      });
      self.clickoutside(canvasContainer, function() {
        block.remove(); 
      });
    }
    
    return canvasContainer;
  },

  makeColorDefaults: function(callback) {
    var self = this;
    var colors = ['transparent', 'rgba(0,0,0,1)', 'rgba(255,255,255,1)'];
    var ul = self.createElement('ul', {
      class: 'color-items'
    });
    [].forEach.call(colors, function(color) {
      var div = self.createElement('div', {
        style: 'background-color:' + color,
        class: 'mask-color',
        color: color
      });
      var li = self.createElement('li', {
        class: color === 'transparent' ? 'color-item transparent' : 'color-item',
      });
      li.appendChild(div);
      ul.appendChild(li);
      li.onclick = function() {
        if(typeof callback === 'function') callback(color);
      }
    });
    return ul;
  },

  initColorPicker: function(canvas, callback) {
    var ctx = canvas.getContext('2d');
    var gradient = ctx.createLinearGradient(0, 0, canvas.width, 0);
    gradient.addColorStop(0, '#ffff00')
    gradient.addColorStop(1 / 6, '#ffff00')
    gradient.addColorStop((1 / 6) * 2, '#00ff00')
    gradient.addColorStop((1 / 6) * 3, '#00ffff')
    gradient.addColorStop((1 / 6) * 4, '#0000ff')
    gradient.addColorStop((1 / 6) * 5, '#ff00ff')
    gradient.addColorStop(1, '#ff0000')
    ctx.fillStyle = gradient
    ctx.fillRect(0, 0, canvas.width, canvas.height)

    gradient = ctx.createLinearGradient(0, 0, 0, canvas.height)
    gradient.addColorStop(0, 'rgba(255, 255, 255, 1)')
    gradient.addColorStop(0.5, 'rgba(255, 255, 255, 0)')
    gradient.addColorStop(1, 'rgba(255, 255, 255, 0)')
    ctx.fillStyle = gradient
    ctx.fillRect(0, 0, canvas.width, canvas.height)

    gradient = ctx.createLinearGradient(0, 0, 0, canvas.height)
    gradient.addColorStop(0, 'rgba(0, 0, 0, 0)')
    gradient.addColorStop(0.5, 'rgba(0, 0, 0, 0)')
    gradient.addColorStop(1, 'rgba(0, 0, 0, 1)')
    ctx.fillStyle = gradient
    ctx.fillRect(0, 0, canvas.width, canvas.height)
    canvas.onclick = function(e) {
      var imgData = ctx.getImageData((e.offsetX / canvas.clientWidth) * canvas.width, (e.offsetY / canvas.clientHeight) * canvas.height, 1, 1)
      var rgba = imgData.data;
      var color = "rgba(" + rgba[0] + ", " + rgba[1] + ", " + rgba[2] + ", " + rgba[3] + ")";
      canvas.remove();
      if(typeof callback === 'function') callback(color);
    };
  },

  isSvgFromStringUrl: function(str) {
    var mimeType = str.match(/[^:]\w+\/[\w-+\d.]+(?=;|,)/);
    return mimeType && mimeType[0] === 'image/svg+xml';
  }
};